"""Utility functions for SCD type 1 and type 2."""
import os
import uuid
from datetime import datetime
from typing import List

import pyspark.sql.functions as F
from delta import DeltaTable
from pyspark.sql import DataFrame as SparkDataFrame
from pyspark.sql import SparkSession
from pyspark.sql.column import Column
from pyspark.sql.types import StringType
from kedro_sumz.utils import get_catalog

TIME_ZONE = "America/Bogota"


def get_current_timestamp(time_zone: str = TIME_ZONE) -> Column:
    """Get current timestamp"""
    return F.from_utc_timestamp(F.current_timestamp(), time_zone)


@F.udf(returnType=StringType())
def create_uuid():
    """Create uuid udf"""
    return str(uuid.uuid4())


@F.udf(returnType=StringType())
def create_static_uuid(value: str):
    """Create uuid udf"""
    return str(uuid.uuid5(uuid.NAMESPACE_OID, value))


def incremental_audit_cols(
    source: SparkDataFrame,
):
    """Add audit columns for incremental loads
    Args:
        source: dataframe to add columns
    Returns:
        a spark dataframe with the required columns
    """
    source = source.select(
        F.col("*"),
        get_current_timestamp().alias("FECHA_CREACION_LAGO_DATOS"),
        get_current_timestamp().alias("FECHA_MODIFICACION_LAGO_DATOS"),
    )
    return source


def partial_incremental_overwrites(
    source: SparkDataFrame,
    target: DeltaTable,
    condition: str,
    target_catalog: str,
    add_new_cols: bool = False,
):
    """Perform a partial incremental overwrite of a DeltaTable
    Args:
        source: dataframe to overwrite
        target: DeltaTable to overwrite
        condition: condition to use for the overwrite
        target_catalog: catalog name of the target table
        add_new_cols: if True, add new columns to the target table
    """
    catalog = get_catalog()
    # pylint: disable=protected-access
    target_catalog = catalog._get_dataset(target_catalog)
    # pylint: disable=protected-access
    target_catalog._save_args["replaceWhere"] = condition
    source = incremental_audit_cols(source.where(F.expr(condition)))
    if not add_new_cols:
        source_cols = [c.upper() for c in source.columns]
        target_cols = [c.upper() for c in target.toDF().columns]
        columns = [c for c in source_cols if c in target_cols]
        source = source.select(columns)
    else:
        # pylint: disable=protected-access
        target_catalog._save_args["mergeSchema"] = "true"
    target_catalog.save(source)


def scd1_audit_cols(
    source: SparkDataFrame,
    key_cols: List[str],
    filter_cols: List = None,
):
    """Add required columns for SCD type 1 first load
    Args:
        source: dataframe to add columns
        key_cols: columns to use as key
        filter_cols: cols to ignore when building the ETLSHA
    Returns:
        a spark dataframe with the required columns
    """
    filter_cols = filter_cols if filter_cols else []

    source_cols = [
        c for c in source.columns if (c not in key_cols) and (c not in filter_cols)
    ]
    source = incremental_audit_cols(source)
    source = source.select(
        F.col("*"),
        F.sha2(
            F.concat(
                *[F.coalesce(F.col(c).cast("String"), F.lit("")) for c in source_cols]
            ),
            256,
        ).alias("ETLSHA"),
    )
    return source


def scd2_audit_cols(
    source: SparkDataFrame,
    key_cols: List[str],
    s_key_name: str = "CLAVE_SUB",
    date_col: str = None,
    gen_uuid: bool = True,
    filter_cols: List = None,
):
    """Add required columns for SCD type 2 first load
    Args:
        source: dataframe to add columns
        key_cols: columns to use as key
        s_key_name: name of the surrogate key column
        date_col: column to use as date
        gen_uuid: if True, generate an uuid column
        filter_cols: cols to ignore when building the ETLSHA
    Returns:
        a spark dataframe with the required columns
    """
    filter_cols = filter_cols if filter_cols else []

    source_cols = [
        c for c in source.columns if (c not in key_cols) and (c not in filter_cols)
    ]
    if gen_uuid:
        source = source.select(create_uuid().alias(s_key_name), F.col("*"))
    source = source.alias("src").select(
        F.col("*"),
        F.coalesce(
            F.col(f"src.{date_col}")
            if date_col and date_col in source.columns
            else F.lit(None),
            F.lit("1900-01-01"),
        )
        .cast("timestamp")
        .alias("FECHA_INICIO_VIGENCIA"),
        F.lit("9999-12-31").cast("timestamp").alias("FECHA_FIN_VIGENCIA"),
        F.lit(True).alias("REGISTRO_ACTIVO"),
        get_current_timestamp().alias("FECHA_CREACION_LAGO_DATOS"),
        get_current_timestamp().alias("FECHA_MODIFICACION_LAGO_DATOS"),
        F.sha2(
            F.concat(
                *[F.coalesce(F.col(c).cast("String"), F.lit("")) for c in source_cols]
            ),
            256,
        ).alias("ETLSHA"),
    )
    return source


def scd1_delta_write(
    source: SparkDataFrame,
    target: DeltaTable,
    key_cols: List[str],
    filter_cols: List[str] = None,
    skip_update_cols: List[str] = None,
    add_new_cols: bool = False,
    partition_filter: str = None,
    delete_not_in_source=True,
):
    """Write a dataframe to a delta table using SCD type 1
    Args:
        source: dataframe to write
        target: delta table to write to
        key_cols: columns to use as key
        filter_cols: cols to ignore when building the ETLSHA
        skip_update_cols: cols to ignore when updating
        add_new_cols: if True, add new columns to the target table
        partition_filter: partition filter to use
        delete_not_in_source: if True, delete rows not in the source table
    """
    if skip_update_cols:
        add_new_cols = False
    else:
        skip_update_cols = []

    filter_cols = filter_cols if filter_cols else []

    source_cols = [
        c for c in source.columns if (c not in key_cols) and (c not in filter_cols)
    ]

    source = scd1_audit_cols(source, key_cols, filter_cols)  # .persist()

    merge_builder = target.alias("tg").merge(
        source.alias("src"),
        (f"({partition_filter}) AND " if partition_filter else "")
        + " AND ".join([f"src.{c}<=>tg.{c}" for c in key_cols]),
    )
    if add_new_cols:
        merge_builder = merge_builder.whenMatchedUpdateAll(
            condition="src.ETLSHA <> tg.ETLSHA"
        ).whenNotMatchedInsertAll()
    else:
        merge_builder = merge_builder.whenMatchedUpdate(
            condition="src.ETLSHA <> tg.ETLSHA",
            set=dict(
                **{
                    c: f"src.{c}"
                    for c in target.toDF().columns
                    if (c in source_cols or c in filter_cols)
                    and (c not in skip_update_cols)
                },
                **{
                    "FECHA_MODIFICACION_LAGO_DATOS": "src.FECHA_MODIFICACION_LAGO_DATOS",
                    "ETLSHA": "src.ETLSHA",
                },
            ),
        ).whenNotMatchedInsert(
            values={
                c: f"src.{c}"
                for c in target.toDF().columns
                if c in source.columns or c in filter_cols
            }
        )
    if delete_not_in_source:
        merge_builder = merge_builder.whenNotMatchedBySourceDelete(
            condition=partition_filter
        )

    merge_builder.execute()


def scd2_delta_write(
    source: SparkDataFrame,
    target: DeltaTable,
    key_cols: List[str],
    s_key_name: str = "CLAVE_SUB",
    date_col: str = None,
    gen_uuid: bool = True,
    filter_cols: List = None,
    partition_cols: List = None,
    add_new_cols: bool = False,
    delete_not_in_source=False,
    partition_filter: str = None,
):
    """Delta SCD type 2 write
    Args:
        source: source dataframe to write
        target: target delta table to write into
        key_cols: columns to use as key
        s_key_name: name of the surrogate key column
        date_col: name of the date column for the merge
        gen_uuid: decide if generate new uuid when inserting
        filter_cols: cols to ignore when building the ETLSHA
        partition_cols: cols to use as partition
        add_new_cols: if True, add new columns to target
        delete_not_in_source: if True, delete records not in source
        partition_filter: partition filter to use
    """
    spark = SparkSession.getActiveSession()

    filter_cols = filter_cols if filter_cols else []

    source_cols = [
        c for c in source.columns if (c not in key_cols) and (c not in filter_cols)
    ]
    if gen_uuid:
        partition_cols = partition_cols if partition_cols else []
        source = source.select(create_uuid().alias(s_key_name), F.col("*"))
        partitions = {}
        keys = {}
        type_map = {
            "string": "NA",
            "date": datetime.strptime("1900-01-01", "%Y-%m-%d").date(),
            "timestamp": datetime.strptime("1900-01-01", "%Y-%m-%d"),
        }
        for col in partition_cols:
            # string, date, timestamp
            type_ = source.select(col).schema[0].jsonValue()["type"]
            partitions[col] = type_map.get(type_, -1)

        for col in key_cols:
            # string, date, timestamp
            type_ = source.select(col).schema[0].jsonValue()["type"]
            keys[col] = type_map.get(type_, -1)

        data = [dict(**{s_key_name: str(uuid.uuid4())}, **keys, **partitions)]
        empty_source = spark.createDataFrame(
            data, schema=target.toDF().select(*source.columns).schema
        )
        source = source.unionByName(empty_source)
    source = source.select(
        F.col("*"),
        F.lit("9999-12-31").cast("timestamp").alias("FECHA_FIN_VIGENCIA"),
        F.lit(True).alias("REGISTRO_ACTIVO"),
        get_current_timestamp().alias("FECHA_CREACION_LAGO_DATOS"),
        get_current_timestamp().alias("FECHA_MODIFICACION_LAGO_DATOS"),
        F.sha2(
            F.concat(
                *[F.coalesce(F.col(c).cast("String"), F.lit("")) for c in source_cols]
            ),
            256,
        ).alias("ETLSHA"),
    )  # .persist()

    audit_cols = [
        "FECHA_INICIO_VIGENCIA",
        "FECHA_FIN_VIGENCIA",
        "REGISTRO_ACTIVO",
        "FECHA_CREACION_LAGO_DATOS",
        "FECHA_MODIFICACION_LAGO_DATOS",
        "ETLSHA",
    ]
    source_cols = [c for c in source.columns if c not in audit_cols] + audit_cols

    updates = (
        source.alias("src")
        .join(
            target.toDF().alias("tg"),
            on=F.expr(
                (f"({partition_filter}) AND " if partition_filter else "") +
                " AND ".join([f"src.{c}<=>tg.{c}" for c in key_cols])
            ),
        )
        .where("tg.REGISTRO_ACTIVO = TRUE AND src.ETLSHA <> tg.ETLSHA")
    )

    staging = updates.select(
        F.lit(0).alias("__key__"),
        F.col("src.*"),
        get_current_timestamp().alias("FECHA_INICIO_VIGENCIA"),
    ).unionByName(
        source.alias("src").select(
            F.lit(1).alias("__key__"),
            F.col("src.*"),
            F.coalesce(
                F.col(f"src.{date_col}")
                if date_col and date_col in source.columns
                else F.lit(None),
                get_current_timestamp(),
            )
            .cast("timestamp")
            .alias("FECHA_INICIO_VIGENCIA"),
        )
    )
    if "DATABRICKS_RUNTIME_VERSION" not in os.environ:
        staging = staging.localCheckpoint(eager=False)


    merge_builder = (
        target.alias("tg")
        .merge(
            staging.alias("st"),
            (f"({partition_filter}) AND " if partition_filter else "") +
            "st.__key__=1 AND " + " AND ".join([f"st.{c}<=>tg.{c}" for c in key_cols]),
        )
        .whenMatchedUpdate(
            condition="tg.REGISTRO_ACTIVO = TRUE AND st.ETLSHA <> tg.ETLSHA",
            set={
                "REGISTRO_ACTIVO": "false",
                "FECHA_FIN_VIGENCIA": get_current_timestamp(),
                "FECHA_MODIFICACION_LAGO_DATOS": "st.FECHA_MODIFICACION_LAGO_DATOS",
            },
        )
    )
    if add_new_cols:
        merge_builder = merge_builder.whenNotMatchedInsert(
            values={c: f"st.{c}" for c in source_cols}
        )
    else:
        merge_builder = merge_builder.whenNotMatchedInsert(
            values={c: f"st.{c}" for c in target.toDF().columns}
        )
    if delete_not_in_source:
        merge_builder = merge_builder.whenNotMatchedBySourceUpdate(
            condition=(f"({partition_filter}) AND " if partition_filter else "") + "tg.REGISTRO_ACTIVO = TRUE",
            set={
                "REGISTRO_ACTIVO": "false",
                "FECHA_FIN_VIGENCIA": get_current_timestamp(),
                "FECHA_MODIFICACION_LAGO_DATOS": get_current_timestamp(),
            },
        )

    merge_builder.execute()



"""Project utilities."""
# pylint: disable=cyclic-import
import importlib
import os
import threading
import traceback
import warnings
from pathlib import Path
from typing import Any, Dict, List, Union

import importlib_resources
import pyspark.sql.functions as F
from pyspark.sql import DataFrame as SparkDataFrame
from pyspark.sql import Window

from kedro.framework.context import KedroContext
from kedro.framework.hooks import _create_hook_manager
from kedro.framework.project import IMPORT_ERROR_MESSAGE, _create_pipeline, settings
from kedro.framework.startup import _get_project_metadata
from kedro.pipeline import Pipeline, pipeline


def _create_kedro_context():
    """
    Initializes the Kedro context.
    """

    project_path = Path(Path.cwd()).resolve()
    metadata = _get_project_metadata(project_path)
    context_class = settings.CONTEXT_CLASS
    env = os.getenv("KEDRO_ENV")
    hook_manager = _create_hook_manager()

    config_loader_class = settings.CONFIG_LOADER_CLASS
    config_loader = config_loader_class(
        conf_source=str(metadata.project_path / settings.CONF_SOURCE),
        env=env,
        **settings.CONFIG_LOADER_ARGS,
    )
    context = context_class(
        package_name=metadata.package_name,
        project_path=metadata.project_path,
        config_loader=config_loader,
        env=env,
        extra_params={},
        hook_manager=hook_manager,
    )
    return context


def get_kedro_context() -> Union[KedroContext, None]:
    """
    Returns the current Kedro context and tries to create one if not possible.
    """
    context = CurrentKedroContext().context
    if context is None:
        try:
            return _create_kedro_context()
        except Exception:  # pylint: disable=broad-except
            return None

    return context


def get_credentials() -> Dict[str, Any]:
    """
    Returns the credentials from the Kedro context.
    """
    if get_kedro_context() is None:
        raise ValueError("Kedro context is not set.")
    # pylint: disable=protected-access
    return get_kedro_context()._get_config_credentials()  # noqa


def get_credential(key: str) -> Union[str, Dict[str, Any], None]:
    """
    Returns the value of a credential from the Kedro context.

    Args:
        key: The key of the credential.

    Returns:
        The value of the credential.
    """
    return get_credentials().get(key, None)


def get_catalog() -> Dict[str, Any]:
    """
    Returns the catalog from the Kedro context.
    """
    if get_kedro_context() is None:
        raise ValueError("Kedro context is not set.")
    # pylint: disable=protected-access
    return get_kedro_context()._get_catalog()  # noqa


class SingletonMeta(type):
    """
    Implementation of a Singleton class using the metaclass method
    """

    _instances = {}
    _lock = threading.Lock()

    def __call__(cls, *args, **kwargs):
        """
        Possible changes to the value of the `__init__` argument do not affect
        the returned instance.
        """
        if cls not in cls._instances:
            with cls._lock:
                if cls not in cls._instances:
                    instance = super(SingletonMeta, cls).__call__(*args, **kwargs)
                    cls._instances[cls] = instance
        return cls._instances[cls]


class CurrentKedroContext(
    metaclass=SingletonMeta
):  # pylint: disable=too-few-public-methods
    """
    A singleton class that stores the current Kedro context.
    """

    def __init__(self, context: KedroContext = None):
        """
        Virtually private constructor.

        Args:
            context: The current Kedro context.
        """
        self._context: KedroContext = context

    @property
    def context(self) -> KedroContext:
        """Returns the current Kedro context."""
        return self._context

    @context.setter
    def context(self, context: KedroContext):
        """Sets the current Kedro context."""
        self._context = context


def find_nested_pipelines(
    base_module: str, package_name: str = "virgo"
) -> Dict[str, Pipeline]:
    """Automatically find modular pipelines having a ``create_pipeline``
    function. By default, projects created using Kedro 0.18.3 and higher
    call this function to autoregister pipelines upon creation/addition.

    Projects that require more fine-grained control can still define the
    pipeline registry without calling this function. Alternatively, they
    can modify the mapping generated by the ``find_pipelines`` function.

    For more information on the pipeline registry and autodiscovery, see
    https://kedro.readthedocs.io/en/latest/nodes_and_pipelines/pipeline_registry.html

    Returns:
        A generated mapping from pipeline names to ``Pipeline`` objects.

    Warns:
        UserWarning: When a module does not expose a ``create_pipeline``
            function, the ``create_pipeline`` function does not return a
            ``Pipeline`` object, or if the module import fails up front.
    """
    pipeline_obj = None
    pipelines_dict = {"_default_": pipeline_obj or pipeline([])}

    # Handle the case that a project doesn't have a pipelines' directory.
    pipelines_package = None
    try:
        pipelines_package = importlib_resources.files(
            f"{package_name}.pipelines.{base_module}"
        )
    except ModuleNotFoundError as exc:
        if str(exc) == f"No module named '{package_name}.pipelines.{base_module}'":
            return pipelines_dict

    for pipeline_dir in pipelines_package.iterdir():
        if not pipeline_dir.is_dir():
            continue

        pipeline_name = pipeline_dir.name
        if pipeline_name == "_pycache_":
            continue

        pipeline_module_name = f"{package_name}.pipelines.{base_module}.{pipeline_name}"
        try:
            pipeline_module = importlib.import_module(pipeline_module_name)
        except:  # pylint: disable=bare-except  # noqa: E722
            warnings.warn(
                IMPORT_ERROR_MESSAGE.format(
                    module=pipeline_module_name, tb_exc=traceback.format_exc()
                )
            )
            continue

        pipeline_obj = _create_pipeline(pipeline_module)
        if pipeline_obj is not None:
            pipelines_dict[pipeline_name] = pipeline_obj
    return pipelines_dict


def fill_data_frame_columns(
    df: SparkDataFrame,  # pylint: disable=invalid-name
    columns_to_fill: List[str],
    method: str = "ffill",
    partition_columns: List[str] = None,
    order_columns: List[str] = None,
    value: Any = None,
) -> SparkDataFrame:
    """Fills the given spark DataFrame columns.

    Fills the given spark DataFrame columns using the specified filling method.

    Args:
        df: The Spark DataFrame containing the columns to be filled.
        columns_to_fill: The list of the names of the columns to be filled.
        method: The filling method.
            It can be one of:
            - ffill: forward filling
            - bfill: backwards filling
            - any: fill with value
        partition_columns: Partition columns used as the base for the filling.
            It can be empty.
        order_columns: Columns used for the ordering when filling. They are only used
            ffill or bfill are used.
        value: Value used to fill the specified columns. It can be a dictionary to
            fill every column with a different value.
    Returns:
        A spark DataFrame with the filled columns.
    Raises:
        RuntimeError: Incorrect function call.
    """
    # pylint: disable=too-many-arguments

    partition_columns = [] if partition_columns is None else partition_columns
    order_columns = [] if order_columns is None else order_columns

    if df is None or columns_to_fill is None:
        raise RuntimeError("One of the required arguments doesn't exist.")

    df_columns = df.columns
    columns_to_retain = [c for c in df_columns if c not in columns_to_fill]
    if value is not None:
        if isinstance(value, dict):
            filled_columns = [
                F.when(F.col(f"`{c}`").isNull(), value[c]).otherwise(F.col(f"`{c}`"))
                for c in columns_to_fill
            ]
        else:
            filled_columns = [
                F.when(F.col(f"`{c}`").isNull(), value).otherwise(F.col(f"`{c}`"))
                for c in columns_to_fill
            ]
        remain_columns = [F.col(f"`{c}`") for c in columns_to_retain]
        return df.select(*remain_columns, *filled_columns).select(
            [f"`{c}`" for c in df_columns]
        )

    if method == "ffill":
        if not order_columns:
            raise RuntimeError("Ordering is required to by ffill.")
        if not partition_columns:
            partition_columns = F.lit(0)
        window_ffill = (
            Window.partitionBy(partition_columns)
            .orderBy(*[F.asc(f"`{c}`") for c in order_columns])
            .rowsBetween(Window.unboundedPreceding, 0)
        )
        filled_columns = [
            F.last(f"`{c}`", ignorenulls=True).over(window_ffill).alias(c)
            for c in columns_to_fill
        ]
        remain_columns = [F.col(f"`{c}`") for c in columns_to_retain]
        return df.select(*remain_columns, *filled_columns).select(
            [f"`{c}`" for c in df_columns]
        )
    if method == "bfill":
        if not order_columns:
            raise RuntimeError("Ordering is required to by bfill.")
        if not partition_columns:
            partition_columns = F.lit(0)
        window_bfill = (
            Window.partitionBy(partition_columns)
            .orderBy(*[F.asc(f"`{c}`") for c in order_columns])
            .rowsBetween(0, Window.unboundedFollowing)
        )
        filled_columns = [
            F.first(f"`{c}`", ignorenulls=True).over(window_bfill).alias(c)
            for c in columns_to_fill
        ]
        remain_columns = [F.col(f"`{c}`") for c in columns_to_retain]
        return df.select(*remain_columns, *filled_columns).select(
            [f"`{c}`" for c in df_columns]
        )
    raise RuntimeError("Unknown method type.")



"""Dummy dataset for testing purposes."""
from typing import Any

from kedro.io import MemoryDataSet


class DummyDataset(MemoryDataSet):
    """Dummy dataset for testing purposes."""

    # pylint: disable=useless-super-delegation,too-few-public-methods
    def __init__(self, data: Any = True, copy_mode: str = None):
        super().__init__(data, copy_mode)



"""``AbstractDataSet`` implementation to access DeltaTables using
``delta-spark``
"""
from pathlib import PurePosixPath
from typing import NoReturn

from delta.tables import DeltaTable

# pylint: disable=unused-import
from kedro.extras.datasets.spark import SparkHiveDataSet  # noqa: F401
from kedro.extras.datasets.spark.spark_dataset import (
    _split_filepath,
    _strip_dbfs_prefix,
)
from kedro.io.core import AbstractDataSet, DataSetError
from pyspark.sql import SparkSession
from pyspark.sql.utils import AnalysisException

from kedro_sumz.utils import get_kedro_context


class DeltaTableDataSet(AbstractDataSet[None, DeltaTable]):
    """``DeltaTableDataSet`` loads data into DeltaTable objects.

    Example usage for the
    `YAML API <https://kedro.readthedocs.io/en/stable/data/\
    data_catalog.html#use-the-data-catalog-with-the-yaml-api>`_:

    .. code-block:: yaml

        weather@spark:
          type: spark.SparkDataSet
          filepath: data/02_intermediate/data.parquet
          file_format: "delta"

        weather@delta:
          type: virgo.extras.datasets.DeltaTableDataSet
          directory: data/02_intermediate
          table: "data.parquet"

        weather_table@delta:
          type: virgo.extras.datasets.delta.DeltaTableDataSet
          directory: intermediate
          table: data

    Example usage for the
    `Python API <https://kedro.readthedocs.io/en/stable/data/\
    data_catalog.html#use-the-data-catalog-with-the-code-api>`_:
    ::

        >>> from pyspark.sql import SparkSession
        >>> from pyspark.sql.types import (StructField, StringType,
        >>>                                IntegerType, StructType)
        >>>
        >>> from kedro.extras.datasets.spark import DeltaTableDataSet, SparkDataSet
        >>>
        >>> schema = StructType([StructField("name", StringType(), True),
        >>>                      StructField("age", IntegerType(), True)])
        >>>
        >>> data = [('Alex', 31), ('Bob', 12), ('Clarke', 65), ('Dave', 29)]
        >>>
        >>> spark_df = SparkSession.builder.getOrCreate().createDataFrame(data, schema)
        >>>
        >>> data_set = SparkHiveDataSet(database="test_database", table="test_table",
        >>>                             write_mode="overwrite")
        >>> data_set.save(spark_df)
        >>> delta_dataset = DeltaTableDataSet(directory="test_database", table="test_table")
        >>> delta_table = delta_dataset.load()
        >>>
        >>> delta_table.update()
    """

    # this dataset cannot be used with ``ParallelRunner``,
    # therefore it has the attribute ``_SINGLE_PROCESS = True``
    # for parallelism within a Spark pipeline please consider
    # using ``ThreadRunner`` instead
    _SINGLE_PROCESS = True

    def __init__(
        self,
        directory: str,
        table: str,
        dataset_type: str = None,
    ) -> None:
        """Creates a new instance of ``DeltaTableDataSet``.

        Args:
            directory: Base path to a folder or a database.
            table: Name of the table or file.
            dataset_type: Type of the dataset. Can be either "table" or "file".
        """

        if not dataset_type and ("/" in directory or "/" in table):
            self._type = "file"
        elif not dataset_type and "." in directory:
            self._type = "table"
        else:
            self._type = dataset_type
        if not self._type:
            raise DataSetError(
                "Could not determine dataset type. Please specify it explicitly."
            )
        if self._type == "table":
            path = f"{directory}.{table}"
            self._database = ".".join(path.split(".")[:-1])
            self._table = table
            self._full_table_address = f"{self._database}.{self._table}"
        else:
            path = f"{directory}/{table}"
            fs_prefix, filepath = _split_filepath(path)
            self._fs_prefix = fs_prefix
            self._filepath = PurePosixPath(filepath)
            self._full_table_address = (
                f"delta.`{self._fs_prefix + str(self._filepath)}`"
            )
            self._type = "file"

    @staticmethod
    def _get_spark() -> SparkSession:
        """Get the active Spark session."""
        if hasattr(get_kedro_context(), "spark_session"):
            return get_kedro_context().spark_session
        else:
            SparkSession.getActiveSession()

    def _load(self) -> DeltaTable:
        return DeltaTable.forName(self._get_spark(), self._full_table_address)

    def _save(self, data: None) -> NoReturn:
        raise DataSetError(f"{self.__class__.__name__} is a read only dataset type")

    def _exists(self) -> bool:
        # noqa # pylint:disable=protected-access
        if self._type == "table":
            return (
                self._get_spark()
                ._jsparkSession.catalog()
                .tableExists(self._database, self._table)
            )

        load_path = _strip_dbfs_prefix(self._fs_prefix + str(self._filepath))

        try:
            self._get_spark().read.load(path=load_path, format="delta")
        except AnalysisException as exception:
            if "is not a Delta table" in exception.desc:
                return False
            raise

        return True

    def _describe(self):
        return {
            "full_table_address": self._full_table_address,
        }



"""Context hooks for the Virgo project."""
from typing import Any, Dict, List, Callable

from kedro.framework.hooks import hook_impl
from kedro.io import DataCatalog
from kedro_sumz.extras.context import KedroSparkContext
from kedro_sumz.utils import CurrentKedroContext
try:
    from kedro_sumz.extras.datasets.mlflow import (
        MlflowMetricsDictDataSet
    )
except ImportError:
    pass


class ContextHooks:  # pylint: disable=too-few-public-methods
    """Context hooks."""

    def __init__(self, callbacks: List[Callable[["KedroSparkContext"], None]] = None):
        callbacks = callbacks if callbacks else []
        self._callbacks = callbacks

    @hook_impl
    def after_context_created(self, context: "KedroSparkContext") -> None:
        """Stores the current kedro context inside a singleton class.

        Args:
            context: The current Kedro context.
        """
        _ = CurrentKedroContext(context)
        for c in self._callbacks:
            c(context)


class MLFLowHooks:  # pylint: disable=too-few-public-methods
    """MLFLow hooks."""

    @hook_impl
    def after_catalog_created(
        self,
        catalog: DataCatalog,
        conf_catalog: Dict[str, Any],
        conf_creds: Dict[str, Any],
        feed_dict: Dict[str, Any],
        save_version: str,
        load_versions: str,
    ):
        """Hook to add the mlflow dataset to the catalog."""
        # we use this hooks to modif "MlflowMetricsDictDataSet" to ensure consistency
        # of the metric name with the catalog name
        for name, dataset in catalog._data_sets.items():
            if isinstance(dataset, MlflowMetricsDictDataSet) and dataset._prefix is None:
                if dataset._run_id is not None:
                    catalog._data_sets[name] = MlflowMetricsDictDataSet(
                        run_id=dataset._run_id, prefix=name
                    )
                else:
                    catalog._data_sets[name] = MlflowMetricsDictDataSet(prefix=name)



"""Custom Kedro-Virgo context class."""
import logging
from pathlib import Path
from typing import Any, Dict, Union

from kedro.config import ConfigLoader
from kedro.framework.context import KedroContext
from pluggy import PluginManager

logger = logging.getLogger(__name__)


class KedroSparkContext(KedroContext):
    """Custom kedro context which initializes the Spark Session."""

    def __init__(
        self,
        package_name: str,
        project_path: Union[str, Path],
        config_loader: ConfigLoader,
        hook_manager: PluginManager,
        env: str = None,
        extra_params: Dict[str, Any] = None,
    ):
        logger.info("Creating Kedro Context")
        super().__init__(
            package_name, project_path, config_loader, hook_manager, env, extra_params
        )
        self._spark_session = None
        self._package_name = (
            package_name if package_name else Path(__file__).parent.name
        )

        if not self._init_databricks():
            self._init_spark()

    def _init_spark(self):
        from pyspark import SparkConf
        from pyspark.sql import SparkSession
        
        if SparkSession.getActiveSession() is None:
            logger.info("Initializing Spark session")
            parameters = self._config_loader["spark"]
            logger.info(parameters)
            spark_conf = SparkConf().setAll(list(parameters.items()))

            spark_session_conf = (
                SparkSession.builder.appName(self._package_name)
                .master("local[*, 4]")
                .config(conf=spark_conf)
            )

            self._spark_session = spark_session_conf.getOrCreate()
            self._spark_session.sparkContext.setLogLevel("WARN")
            logger.info("Spark Web URL: %s", self._spark_session.sparkContext.uiWebUrl)
        else:
            self._spark_session = SparkSession.getActiveSession()

    def _init_databricks(self) -> bool:
        try:
            from databricks.connect import DatabricksSession
            from pyspark.sql import SparkSession

            if SparkSession.getActiveSession() is None:
                logger.info("Initializing Databricks session")
                self._spark_session = DatabricksSession.builder.profile(
                    "connect"
                ).getOrCreate()
            else:
                self._spark_session = SparkSession.getActiveSession()
            return True
        except ImportError:
            logger.info(
                "Databricks Connect is not installed. Initializing Spark session instead."
            )
            return False

    @property
    def spark_session(self):
        """Spark session property."""
        return self._spark_session



"""``AbstractDataSet`` implementation to access Spark dataframes using
``pyspark`` on Apache Hive.
"""
import importlib
import pickle
from copy import deepcopy
from pathlib import PurePosixPath
from typing import Any, Dict
from warnings import warn

from kedro.extras.datasets.spark.spark_dataset import (
    _split_filepath,
    _strip_dbfs_prefix,
)
from kedro.io.core import (
    AbstractDataSet,
    DataSetError,
)
from pyspark.ml import PipelineModel

from pyspark.ml.util import GeneralJavaMLWritable
from pyspark.sql import DataFrame, SparkSession

from pyspark.sql.utils import AnalysisException
from s3fs import S3FileSystem


# pylint:disable=too-many-instance-attributes
class SparkRegressionModelDataSet(AbstractDataSet):
    """``SparkRegressionModelDataSet`` loads and saves spark ml regression models."""

    _SINGLE_PROCESS = True
    DEFAULT_SAVE_ARGS = {}  # type: Dict[str, Any]
    DEFAULT_LOAD_ARGS = {}  # type: Dict[str, Any]

    # pylint:disable=too-many-arguments
    def __init__(
        self,
        filepath: str,
        credentials: Dict[str, Any] = None,
        load_args: Dict[str, Any] = None,
        save_args: Dict[str, Any] = None,
    ) -> None:
        """Creates a new instance of ``SparkRegressionModelDataSet``.

        Args:
            filepath: Path to the saved model.
            credentials: Credentials required to get access to the filesystem.
            dataset_type: Type of the dataset. Can be either "table" or "file".
            load_args: Load args passed to Spark LinearRegressionModel
            save_args: Optional mapping of any options,
                passed to the LinearRegressionModel save method.

        Raises:
            DataSetError: Invalid configuration supplied
        """
        credentials = deepcopy(credentials) or {}
        fs_prefix, filepath = _split_filepath(filepath)
        if fs_prefix in ("s3a://", "s3n://"):
            if fs_prefix == "s3n://":
                warn(
                    "`s3n` filesystem has now been deprecated by Spark, "
                    "please consider switching to `s3a`",
                    DeprecationWarning,
                )
            _s3 = S3FileSystem(**credentials)
            path = PurePosixPath(filepath)
        else:
            path = PurePosixPath(filepath)

        self._filepath = path
        self._fs_prefix = fs_prefix
        # Handle default load and save arguments
        self._load_args = deepcopy(self.DEFAULT_LOAD_ARGS)
        if load_args is not None:
            self._load_args.update(load_args)
        self._save_args = deepcopy(self.DEFAULT_SAVE_ARGS)
        if save_args is not None:
            self._save_args.update(save_args)

        self._supported_modes = {
            "overwrite",
            "error",
            "errorifexists",
            "ignore",
        }
        self._write_mode = self._save_args.pop("mode", "errorifexists")
        self._handle_write_mode()

    def _describe(self) -> Dict[str, Any]:
        return {
            "filepath": self._fs_prefix + str(self._filepath),
            "load_args": self._load_args,
            "save_args": self._save_args,
        }

    @staticmethod
    def _get_spark() -> SparkSession:
        """Get the active Spark session."""
        return SparkSession.getActiveSession()

    def _exists(self) -> bool:
        try:
            self._load()
        except AnalysisException as exception:
            if (
                exception.desc.startswith("Path does not exist:")
                or "is not a Delta table" in exception.desc
            ):
                return False
            raise
        return True

    def _load(self) -> DataFrame:
        load_path = _strip_dbfs_prefix(self._fs_prefix + str(self._filepath))
        regression_module = importlib.import_module("pyspark.ml.regression")
        cls = getattr(
            regression_module,
            self._load_args.get("regression_model", "LinearRegressionModel"),
        )
        return cls.load(load_path)

    def _save(self, data: GeneralJavaMLWritable) -> None:
        save_path = _strip_dbfs_prefix(self._fs_prefix + str(self._filepath))
        if self._write_mode == "overwrite":
            data.write().overwrite().save(save_path)
        elif self._write_mode == "error":
            raise DataSetError(
                f"It is not possible to perform 'save()' "
                f"with mode '{self._write_mode}' on 'SparkModelDataSet'. "
            )
        else:
            data.write().save(save_path)

    def _handle_write_mode(self) -> None:
        if self._write_mode and self._write_mode not in self._supported_modes:
            raise DataSetError(
                f"It is not possible to perform 'save()' "
                f"with mode '{self._write_mode}' on 'SparkModelDataSet'. "
            )

    def __getstate__(self) -> None:
        raise pickle.PicklingError(
            "PySpark datasets objects cannot be pickled "
            "or serialised as Python objects."
        )


# pylint:disable=too-many-instance-attributes
class SparkPipelineModelDataSet(AbstractDataSet):
    """``SparkPipelineModelDataSet`` loads and saves spark ml pipeline models."""

    _SINGLE_PROCESS = True
    DEFAULT_SAVE_ARGS = {}  # type: Dict[str, Any]
    DEFAULT_LOAD_ARGS = {}  # type: Dict[str, Any]

    # pylint:disable=too-many-arguments
    def __init__(
        self,
        filepath: str,
        credentials: Dict[str, Any] = None,
        load_args: Dict[str, Any] = None,
        save_args: Dict[str, Any] = None,
    ) -> None:
        """Creates a new instance of ``SparkPipelineModelDataSet``.

        Args:
            filepath: Path to the saved model.
            credentials: Credentials required to get access to the filesystem.
            dataset_type: Type of the dataset. Can be either "table" or "file".
            load_args: Load args passed to Spark LinearRegressionModel
            save_args: Optional mapping of any options,
                passed to the LinearRegressionModel save method.

        Raises:
            DataSetError: Invalid configuration supplied
        """
        credentials = deepcopy(credentials) or {}
        fs_prefix, filepath = _split_filepath(filepath)
        if fs_prefix in ("s3a://", "s3n://"):
            if fs_prefix == "s3n://":
                warn(
                    "`s3n` filesystem has now been deprecated by Spark, "
                    "please consider switching to `s3a`",
                    DeprecationWarning,
                )
            _s3 = S3FileSystem(**credentials)
            path = PurePosixPath(filepath)
        else:
            path = PurePosixPath(filepath)

        self._filepath = path
        self._fs_prefix = fs_prefix
        # Handle default load and save arguments
        self._load_args = deepcopy(self.DEFAULT_LOAD_ARGS)
        if load_args is not None:
            self._load_args.update(load_args)
        self._save_args = deepcopy(self.DEFAULT_SAVE_ARGS)
        if save_args is not None:
            self._save_args.update(save_args)

        self._supported_modes = {
            "overwrite",
            "error",
            "errorifexists",
            "ignore",
        }
        self._write_mode = self._save_args.pop("mode", "errorifexists")
        self._handle_write_mode()

    def _describe(self) -> Dict[str, Any]:
        return {
            "filepath": self._fs_prefix + str(self._filepath),
            "load_args": self._load_args,
            "save_args": self._save_args,
        }

    @staticmethod
    def _get_spark() -> SparkSession:
        """Get the active Spark session."""
        return SparkSession.getActiveSession()

    def _exists(self) -> bool:
        try:
            self._load()
        except AnalysisException as exception:
            if (
                exception.desc.startswith("Path does not exist:")
                or "is not a Delta table" in exception.desc
            ):
                return False
            raise
        return True

    def _load(self) -> DataFrame:
        load_path = _strip_dbfs_prefix(self._fs_prefix + str(self._filepath))
        return PipelineModel.load(load_path)

    def _save(self, data: GeneralJavaMLWritable) -> None:
        save_path = _strip_dbfs_prefix(self._fs_prefix + str(self._filepath))
        if self._write_mode == "overwrite":
            data.write().overwrite().save(save_path)
        elif self._write_mode == "error":
            raise DataSetError(
                f"It is not possible to perform 'save()' "
                f"with mode '{self._write_mode}' on 'SparkPipelineModelDataSet'. "
            )
        else:
            data.write().save(save_path)

    def _handle_write_mode(self) -> None:
        if self._write_mode and self._write_mode not in self._supported_modes:
            raise DataSetError(
                f"It is not possible to perform 'save()' "
                f"with mode '{self._write_mode}' on 'SparkPipelineModelDataSet'. "
            )

    def __getstate__(self) -> None:
        raise pickle.PicklingError(
            "PySpark datasets objects cannot be pickled "
            "or serialised as Python objects."
        )



"""``AbstractDataSet`` implementation to access Spark dataframes using
``pyspark`` on Apache Hive.
"""
import json
import pickle
from copy import deepcopy
from pathlib import PurePosixPath
from typing import Any, Dict

import fsspec
from kedro.extras.datasets.spark.spark_dataset import (
    _split_filepath,
)
from kedro.io.core import (
    AbstractDataSet,
    DataSetError,
    get_filepath_str,
    get_protocol_and_path,
)
from pyspark.sql import DataFrame, SparkSession
from pyspark.sql.types import StructType
from pyspark.sql.utils import AnalysisException

from kedro_sumz.utils import get_kedro_context


# pylint:disable=too-many-instance-attributes
class SparkTableDataSet(AbstractDataSet[DataFrame, DataFrame]):
    """``SparkTableDataSet`` loads and saves Spark dataframes.
    This dataset is able to load tables from Hive and save them back to Hive.

    This DataSet has some key assumptions:

    - Tables are identified by database and table name if the file
        path contains does not contain a slash

    Example usage for the
    `YAML API <https://kedro.readthedocs.io/en/stable/data/\
    data_catalog.html#use-the-data-catalog-with-the-yaml-api>`_:

    .. code-block:: yaml

        spark_dataset:
          type: scorpius.extras.datasets.spark.SparkTableDataSet
          directory: hive_database
          table: table_name
          save_args:
            mode: overwrite

    Example usage for the
    `Python API <https://kedro.readthedocs.io/en/stable/data/\
    data_catalog.html#use-the-data-catalog-with-the-code-api>`_:
    ::

        >>> from pyspark.sql import SparkSession
        >>> from pyspark.sql.types import (StructField, StringType,
        >>>                                IntegerType, StructType)
        >>>
        >>> from kedro_sumz.extras.datasets.spark import SparkTableDataSet
        >>>
        >>> schema = StructType([StructField("name", StringType(), True),
        >>>                      StructField("age", IntegerType(), True)])
        >>>
        >>> data = [('Alex', 31), ('Bob', 12), ('Clarke', 65), ('Dave', 29)]
        >>>
        >>> spark_df = SparkSession.builder.getOrCreate().createDataFrame(data, schema)
        >>>
        >>> data_set = SparkTableDataSet(directory="test_database", table="test_table")
        >>> data_set.save(spark_df)
        >>> reloaded = data_set.load()
        >>>
        >>> reloaded.take(4)
    """

    _SINGLE_PROCESS = True
    DEFAULT_SAVE_ARGS = {}  # type: Dict[str, Any]
    DEFAULT_LOAD_ARGS = {}  # type: Dict[str, Any]

    # pylint:disable=too-many-arguments
    def __init__(
        self,
        directory: str,
        table: str,
        table_format: str = "delta",
        dataset_type: str = None,
        load_args: Dict[str, Any] = None,
        save_args: Dict[str, Any] = None,
    ) -> None:
        """Creates a new instance of ``SparkHiveDataSet``.

        Args:
            directory: Base path to a folder or a database.
            table: Name of the table or file.
            table_format: File format used during load and save
                operations. These are formats supported by the running
                SparkContext include parquet, csv, delta. For a list of supported
                formats please refer to Apache Spark documentation at
                https://spark.apache.org/docs/latest/sql-programming-guide.html
            dataset_type: Type of the dataset. Can be either "table" or "file".
            load_args: Load args passed to Spark DataFrameReader load method.
                It is dependent on the selected file format. You can find
                a list of read options for each supported format
                in Spark DataFrame read documentation:
                https://spark.apache.org/docs/latest/api/python/getting_started/quickstart_df.html
            save_args: Optional mapping of any options,
                passed to the `DataFrameWriter.saveAsTable` as kwargs.
                Key example of this is `partitionBy` which allows data partitioning
                on a list of column names.
                Other `HiveOptions` can be found here:
                https://spark.apache.org/docs/latest/sql-data-sources-hive-tables.html#specifying-storage-format-for-hive-tables

        Raises:
            DataSetError: Invalid configuration supplied
        """
        if not dataset_type and ("/" in directory or "/" in table):
            self._type = "file"
        elif not dataset_type and "." in directory:
            self._type = "table"
        else:
            self._type = dataset_type
        if not self._type:
            raise DataSetError(
                "Could not determine dataset type. Please specify it explicitly."
            )
        if self._type == "table":
            directory = directory.rstrip(".")
            table = table.lstrip(".")
            path = f"{directory}.{table}"
            self._database = ".".join(path.split(".")[:-1])
            self._table = table
            self._full_table_address = f"{self._database}.{self._table}"
        else:
            directory = directory.rstrip("/")
            table = table.lstrip("/")
            path = (
                f"{directory}.{table}"
                if table_format == "deltasharing"
                else f"{directory}/{table}"
            )
            fs_prefix, filepath = _split_filepath(path)
            self._fs_prefix = fs_prefix
            self._filepath = PurePosixPath(filepath)
            self._type = "file"

        # Handle default load and save arguments
        self._load_args = deepcopy(self.DEFAULT_LOAD_ARGS)
        if load_args is not None:
            self._load_args.update(load_args)
        self._save_args = deepcopy(self.DEFAULT_SAVE_ARGS)
        if save_args is not None:
            self._save_args.update(save_args)

        self._supported_modes = {
            "append",
            "overwrite",
            "error",
            "errorifexists",
            "ignore",
        }
        self._write_mode = self._save_args.pop("mode", "errorifexists")
        self._write_mode = (
            "error" if table_format == "deltasharing" else self._write_mode
        )

        # Handle schema load argument
        self._schema = self._load_args.pop("schema", None)
        if self._type == "file" and self._schema is not None:
            if isinstance(self._schema, dict):
                self._schema = self._load_schema_from_file(self._schema)

        self._file_format = table_format
        self._handle_write_mode()

    @staticmethod
    def _load_schema_from_file(schema: Dict[str, Any]) -> StructType:

        filepath = schema.get("filepath")
        if not filepath:
            raise DataSetError(
                "Schema load argument does not specify a 'filepath' attribute. Please"
                "include a path to a JSON-serialised 'pyspark.sql.types.StructType'."
            )

        credentials = deepcopy(schema.get("credentials")) or {}
        protocol, schema_path = get_protocol_and_path(filepath)
        file_system = fsspec.filesystem(protocol, **credentials)
        pure_posix_path = PurePosixPath(schema_path)
        load_path = get_filepath_str(pure_posix_path, protocol)

        # Open schema file
        with file_system.open(load_path) as fs_file:

            try:
                return StructType.fromJson(json.loads(fs_file.read()))
            except Exception as exc:
                raise DataSetError(
                    f"Contents of 'schema.filepath' ({schema_path}) are invalid. Please"
                    f"provide a valid JSON-serialised 'pyspark.sql.types.StructType'."
                ) from exc

    def _describe(self) -> Dict[str, Any]:
        if self._type == "file":
            return {
                "filepath": self._fs_prefix + str(self._filepath),
                "file_format": self._file_format,
                "load_args": self._load_args,
                "save_args": self._save_args,
            }

        return {
            "database": self._database,
            "table": self._table,
            "partition": self._save_args.get("partitionBy"),
            "format": self._file_format,
            "load_args": self._load_args,
            "save_args": self._save_args,
        }

    @staticmethod
    def _get_spark() -> SparkSession:
        """Get the active Spark session."""
        if hasattr(get_kedro_context(), "spark_session"):
            return get_kedro_context().spark_session
        else:
            SparkSession.getActiveSession()

    def _create_hive_table(self, data: DataFrame, mode: str = None):
        _mode: str = mode or self._write_mode
        data.write.saveAsTable(
            self._full_table_address,
            mode=_mode,
            format=self._file_format,
            **self._save_args,
        )

    def _load(self) -> DataFrame:
        read_obj = self._get_spark().read
        if self._type == "file":
            load_path = self._fs_prefix + str(self._filepath)
            # Pass schema if defined
            if self._schema:
                read_obj = read_obj.schema(self._schema)

            return read_obj.load(load_path, self._file_format, **self._load_args)

        return read_obj.table(self._full_table_address)

    def _save(self, data: DataFrame) -> None:
        if self._type == "file":
            save_path = self._fs_prefix + str(self._filepath)
            data.write.save(
                save_path, self._file_format, mode=self._write_mode, **self._save_args
            )
        else:
            self._validate_save(data)
            self._create_hive_table(data=data)

    def _validate_save(self, data: DataFrame):
        # do not validate when the table doesn't exist
        # or if the `write_mode` is set to overwrite
        if (
            (self._type != "table")
            or (self._write_mode == "overwrite")
            or ("mergeSchema" in self._save_args)
            or (not self._exists())
        ):
            return
        hive_dtypes = set(self._load().dtypes)
        data_dtypes = set(data.dtypes)
        if data_dtypes != hive_dtypes:
            new_cols = data_dtypes - hive_dtypes
            missing_cols = hive_dtypes - data_dtypes
            raise DataSetError(
                f"Dataset does not match hive table schema.\n"
                f"Present on insert only: {sorted(new_cols)}\n"
                f"Present on schema only: {sorted(missing_cols)}"
            )

    def _exists(self) -> bool:
        if self._type == "file":
            load_path = self._fs_prefix + str(self._filepath)

            try:
                self._get_spark().read.load(load_path, self._file_format)
            except AnalysisException as exception:
                if (
                    exception.desc.startswith("Path does not exist:")
                    or "is not a Delta table" in exception.desc
                ):
                    return False
                raise
            return True

        # noqa # pylint:disable=protected-access
        return (
            self._get_spark()
            ._jsparkSession.catalog()
            .tableExists(self._database, self._table)
        )

    def _handle_write_mode(self) -> None:
        if self._write_mode and self._write_mode not in self._supported_modes:
            raise DataSetError(
                f"It is not possible to perform 'save()' for file format "
                f"'{self._file_format}' "
                f"with mode '{self._write_mode}' on 'SparkTableDataSet'. "
            )

    def __getstate__(self) -> None:
        raise pickle.PicklingError(
            "PySpark datasets objects cannot be pickled "
            "or serialised as Python objects."
        )



from functools import partial
from typing import Any, Dict, Optional
from copy import deepcopy
import mlflow
from kedro.io import AbstractDataSet, DataSetError
from mlflow.tracking import MlflowClient
from kedro_mlflow.io.models.mlflow_model_logger_dataset import (
    MlflowModelLoggerDataSet
)
from kedro_sumz.utils.mlflow import ModelWithConfig


class MlflowModelWithConfigDataSet(MlflowModelLoggerDataSet):
    """
    Wrapper for saving, logging and loading for all MLFlow model flavor
    including its configuration
    """
    def __init__(
        self,
        flavor: str,
        run_id: Optional[str] = None,
        artifact_path: Optional[str] = "model",
        pyfunc_workflow: Optional[str] = None,
        load_args: Optional[Dict[str, Any]] = None,
        save_args: Optional[Dict[str, Any]] = None
    ) -> None:
        """Initializes MlflowModelWithConfigDataSet

        Args:
            flavor (str): Flavor of the model
            run_id (Optional[str], optional): Mlflow run id. Defaults to None.
            artifact_path (Optional[str], optional): Path of the model artifact. Defaults to "model".
            pyfunc_workflow (Optional[str], optional): Name of the pyfunc model. Defaults to None.
            load_args (Optional[Dict[str, Any]], optional): Load args used in the 
            mlflow load_model function. Defaults to None.
            save_args (Optional[Dict[str, Any]], optional): Save args used in the
            mlflow log_model function. Defaults to None.
        """
        super().__init__(flavor, run_id, artifact_path, pyfunc_workflow, load_args, save_args)

    def _save_model_in_run(self, model: ModelWithConfig):
        """
        Saves a model in a run as an artifact

        Args:
            model (ModelWithConfig): Object which stores the model and its configuration

        Raises:
            DataSetError: When the object passed is not an instance of the ModelWithConfig class
        """
        if not isinstance(model, ModelWithConfig):
            raise DataSetError(
                "The model to be logged must be an instance of the ModelWithConfig class"
            )

        if not self._logging_activated:
            return
        
        save_args = self._prepare_save_args(
            model_with_config=model
        )

        if self._flavor == "mlflow.pyfunc":
            # PyFunc models utilise either `python_model` or `loader_module`
            # workflow. We we assign the passed `model` object to one of those keys
            # depending on the chosen `pyfunc_workflow`.
            save_args[self._pyfunc_workflow] = model.model
            self._mlflow_model_module.log_model(
                **save_args
            )
        else:
            # Otherwise we save using the common workflow
            self._mlflow_model_module.log_model(
                model.model,
                **save_args
            )
    
    def _prepare_save_args(
            self,
            model_with_config: ModelWithConfig
    ) -> Dict[str, Any]:
        save_args = deepcopy(self._save_args)

        save_args["signature"] = self._save_arg_value(
            catalog_save_arg=self._save_args.get("signature"),
            model_with_config_arg=model_with_config.signature
        )

        save_args["conda_env"] = self._save_arg_value(
            catalog_save_arg=self._save_args.get("conda_env"),
            model_with_config_arg=model_with_config.conda_env
        )

        save_args["pip_requirements"] = self._save_arg_value(
            catalog_save_arg=self._save_args.get("pip_requirements"),
            model_with_config_arg=model_with_config.pip_requirements
        )

        save_args["extra_pip_requirements"] = self._save_arg_value(
            catalog_save_arg=self._save_args.get("extra_pip_requirements"),
            model_with_config_arg=model_with_config.extra_pip_requirements
        )

        save_args["artifact_path"] = self._artifact_path

        return save_args
    
    def _save_arg_value(
            self,
            catalog_save_arg: Any,
            model_with_config_arg: Any
    )->Any:
        if catalog_save_arg is None and model_with_config_arg is None:
            return None
        elif catalog_save_arg is None:
            return model_with_config_arg
        else:
            return catalog_save_arg


class MlflowMetricsDictDataSet(AbstractDataSet):
    """This class represent MLflow metrics dictionary dataset."""

    def __init__(
        self,
        prefix: str = None,
        run_id: str = None,
    ):
        """Initialise MlflowMetricsDictDataSet.

        Args:
            prefix (str, optional): Prefix for metrics logged in MLflow. Defaults to None
            run_id (str, optional): ID of MLflow run. Defaults to None
        """
        self._prefix = prefix
        self.run_id = run_id
        self._logging_activated = True  # by default, logging is activated!

    @property
    def run_id(self):
        """Get run id.

        If active run is not found, tries to find last experiment.

        Raise `DataSetError` exception if run id can't be found.

        Returns:
            str: String contains run_id.
        """
        if self._run_id is not None:
            return self._run_id
        run = mlflow.active_run()
        if run:
            return run.info.run_id
        raise DataSetError("Cannot find run id.")

    @run_id.setter
    def run_id(self, run_id):
        self._run_id = run_id

    # we want to be able to turn logging off for an entire pipeline run
    # To avoid that a single call to a dataset in the catalog creates a new run automatically
    # we want to be able to turn everything off
    @property
    def _logging_activated(self):
        return self.__logging_activated

    @_logging_activated.setter
    def _logging_activated(self, flag):
        if not isinstance(flag, bool):
            raise ValueError(f"_logging_activated must be a boolean, got {type(flag)}")
        self.__logging_activated = flag

    def _load(self) -> Dict[str, float]:
        """Load MlflowMetricsDictDataSet.

        Returns:
            Dict[str, float]: Dictionary with MLflow metrics.
        """
        all_metrics = self._get_all_metrics()

        dataset_metrics = self._filter_dataset_metrics(all_metrics)

        if not dataset_metrics:
            raise DataSetError(
            "Tried to load an empty metrics dictionary"
            f"be sure that metrics with the prefix '{self._prefix}' exists"
            "in the current run"
            )
            
        return dataset_metrics

    def _save(self, data: Dict[str, float]) -> None:
        """
        Save given MLflow metrics dictionary dataset 
        and log it in MLflow as metrics.

        Args:
            data (Dict[str, float]): Metrics dictionary.
        """
        if not data:
            raise DataSetError(
                "Metrics dictionary should have at least one metric"
            )

        client = MlflowClient()
        try:
            run_id = self.run_id
        except DataSetError:
            # If run_id can't be found log_metric would create new run.
            run_id = None

        log_metric = (
            partial(client.log_metric, run_id)
            if run_id is not None
            else mlflow.log_metric
        )

        if not all([isinstance(value, float) for value in data.values()]):
            raise DataSetError(f"All metric values should be of type `float`")

        metrics = {f"{self._prefix}.{key}": value for key, value in data.items()}

        if self._logging_activated:
            for k, v in metrics.items():
                log_metric(k, v)

    def _exists(self) -> bool:
        """Check if MLflow metrics dataset exists.

        Returns:
            bool: Is MLflow metrics dataset exists?
        """
        all_metrics = self._get_all_metrics()
        dataset_metrics = self._filter_dataset_metrics(all_metrics)
        return len(dataset_metrics.keys())>0

    def _describe(self) -> Dict[str, Any]:
        """Describe MLflow metrics dataset.

        Returns:
            Dict[str, Any]: Dictionary with MLflow metrics 
            dictionary dataset description.
        """
        return {
            "run_id": self._run_id,
            "prefix": self._prefix,
        }

    def _filter_dataset_metrics(self, metrics: Dict[str, float]) -> Dict[str, float]:
        """
        Filter a metric dictionary by keeping only those that 
        belongs to the dataset

        Args:
            metric (Dict[str, float]): Metrics dictionary.
        Returns:
            Dict[str, float]: Dictionary with the metrics that belongs 
            to the dataset
        """
        dataset_metrics = {
            k: v for k, v in metrics.items()
            if k.startswith(self._prefix)
        }

        return dataset_metrics

    def _get_all_metrics(self)->Dict[str, float]:
        """Retrieve all metrics logged in the current run

        Returns:
            Dict[str, float]: Dictionary with all the metrics logged
            in the current run
        """
        client = MlflowClient()
        run = client.get_run(self.run_id)
        return run.data.metrics



from typing import Any, Dict, List, Optional
from mlflow.models.signature import ModelSignature
from mlflow.models.signature import infer_signature
from pyspark.sql import DataFrame as SparkDataFrame
from pyspark.sql.types import DecimalType
from pyspark.sql.functions import col
from pyspark.ml.linalg import VectorUDT
from dataclasses import dataclass


def signature_from_examples(
  input_example: Any,
  output_example: Any,
  params_example: Optional[Dict[str, Any]] = None
) -> ModelSignature:
    """
    Creates a Mlflow signature object from the examples
    of a model's input, output or parameters

    Args:
        input_example (Any): Model's input example
        output_example (Any): Model's output example
        params_example (Optional[Dict[str, Any]], optional): Model's parameters example.
        Defaults to None.

    Returns:
        ModelSignature: Mlflow signature object
    """
    if isinstance(input_example, SparkDataFrame):
        input_example = _preprocess_spark_example_data(
            example_data=input_example
        )

    if isinstance(output_example, SparkDataFrame):
        output_example = _preprocess_spark_example_data(
            example_data=output_example
        )

    return infer_signature(
        model_input=input_example,
        model_output=output_example,
        params=params_example
    )


def signature_from_spark_model_predictions(
    model: Any,
    input_example: SparkDataFrame,
    pred_col: str,
    features: Optional[List[str]] = None,
    params_example: Optional[Dict[str, Any]] = None
  ) -> ModelSignature:
    """
    Creates a Mlflow signature object using a model, it's input example
    and its parameters example. It generates the output example of the model
    using it's predictions on the input example supplied.

    Args:
        model (Any): Spark model
        input_example (SparkDataFrame): Model's input example
        pred_col (str): Name of the prediction column
        features (Optional[List[str]], optional): List of the columns used
        as features. Defaults to None.
        params_example (Optional[Dict[str, Any]], optional): Model's parameters
        example. Defaults to None.

    Returns:
        ModelSignature: Mlflow signature object
    """
    input_example = _preprocess_spark_example_data(
        example_data=input_example
    )

    if features:
        input_example = input_example.select(features)

    return infer_signature(
        model_input=input_example,
        model_output=model.transform(
            input_example.limit(1000)
        ).select(pred_col),
        params=params_example
    )


def _preprocess_spark_example_data(
    example_data:SparkDataFrame,
)->SparkDataFrame:
    """
    Preprocess a pyspark dataframe that represents either an input or output example data of a 
    model in order to allow MLFlow to infer its schema correctly.

    Args:
        example_data (SparkDataframe): Example data

    Returns:
        SparkDataframe: Transformed data
    """

    for field in example_data.schema.fields:
        if isinstance(field.dataType, DecimalType):
            example_data = example_data.withColumn(
                field.name,
                col(field.name).cast('double')
            )
        elif isinstance(field.dataType, VectorUDT):
            example_data = example_data.drop(field.name)
    return example_data


@dataclass
class ModelWithConfig:
    """
    Data class to store a model along with its configuration

    Attributes:
        model (Any): Model.
        signature (ModelSignature): A Mlflow signature object.
        conda_env (Dict[str,Any], optional): Conda enviroment configuration 
        dictionary. Defaults to None.
        pip_requirements (Dict[str,Any], optional): Pip requirements configuration 
        dictionary. Defaults to None.
        extra_pip_requirements (Dict[str,Any], optional): Extra pip requirements configuration 
        dictionary. Defaults to None.
    """
    model: Any
    signature: ModelSignature
    conda_env: Optional[Dict[str, Any]] = None
    pip_requirements: Optional[Dict[str, Any]] = None
    extra_pip_requirements: Optional[Dict[str, Any]] = None
